#ifndef ANACAL_FPFS_H
#define ANACAL_FPFS_H

#include "fpfs/base.h"
#include "fpfs/catalog.h"
#include "fpfs/image.h"

namespace anacal {
    void pyExportFpfs(py::module_& m);
}

#endif // ANACAL_FPFS_H
